@@ -24,11 +24,13 @@ const {
   expandIri: _expandIri,
   getContextValue: _getContextValue,
   isKeyword: _isKeyword,
-  process: _processContext
+  process: _processContext,
+  processingMode: _processingMode
 } = require('./context');
 
 const {
-  removeBase: _removeBase
+  removeBase: _removeBase,
+  prependBase: _prependBase
 } = require('./url');
 
 const {
@@ -51,9 +53,9 @@ module.exports = api;
  * @param options the compaction options.
  * @param compactionMap the compaction map to use.
  *
- * @return the compacted value.
+ * @return a promise that resolves to the compacted value.
  */
-api.compact = ({
+api.compact = async ({
   activeCtx,
   activeProperty = null,
   element,
@@ -65,7 +67,7 @@ api.compact = ({
     let rval = [];
     for(let i = 0; i < element.length; ++i) {
       // compact, dropping any null values unless custom mapped
-      let compacted = api.compact({
+      let compacted = await api.compact({
         activeCtx,
         activeProperty,
         element: element[i],
@@ -73,8 +75,7 @@ api.compact = ({
         compactionMap
       });
       if(compacted === null) {
-        // TODO: use `await` to support async
-        compacted = compactionMap({
+        compacted = await compactionMap({
           unmappedValue: element[i],
           activeCtx,
           activeProperty,
@@ -102,11 +103,11 @@ api.compact = ({
   // use any scoped context on activeProperty
   const ctx = _getContextValue(activeCtx, activeProperty, '@context');
   if(!_isUndefined(ctx)) {
-    // Note: spec's `from term` var is named `isPropertyTermScopedContext`
-    activeCtx = _processContext({
+    activeCtx = await _processContext({
       activeCtx,
       localCtx: ctx,
-      isPropertyTermScopedContext: true,
+      propagate: true,
+      overrideProtected: true,
       options
     });
   }
@@ -138,13 +139,48 @@ api.compact = ({
       return rval;
     }
 
+    // if expanded property is @list and we're contained within a list
+    // container, recursively compact this item to an array
+    if(_isList(element)) {
+      const container = _getContextValue(
+        activeCtx, activeProperty, '@container') || [];
+      if(container.includes('@list')) {
+        return api.compact({
+          activeCtx,
+          activeProperty,
+          element: element['@list'],
+          options,
+          compactionMap
+        });
+      }
+    }
+
     // FIXME: avoid misuse of active property as an expanded property?
     const insideReverse = (activeProperty === '@reverse');
 
     const rval = {};
 
-    // revert type scoped context
-    activeCtx = activeCtx.revertTypeScopedContext();
+    // original context before applying property-scoped and local contexts
+    const inputCtx = activeCtx;
+
+    // revert to previous context, if there is one,
+    // and element is not a value object or a node reference
+    if(!_isValue(element) && !_isSubjectReference(element)) {
+      activeCtx = activeCtx.revertToPreviousContext();
+    }
+
+    // apply property-scoped context after reverting term-scoped context
+    const propertyScopedCtx =
+      _getContextValue(inputCtx, activeProperty, '@context');
+    if(!_isUndefined(propertyScopedCtx)) {
+      activeCtx = await _processContext({
+        activeCtx,
+        localCtx: propertyScopedCtx,
+        propagate: true,
+        overrideProtected: true,
+        options
+      });
+    }
 
     if(options.link && '@id' in element) {
       // store linked element
@@ -169,13 +205,13 @@ api.compact = ({
         {activeCtx: typeContext, iri: type, relativeTo: {vocab: true}});
 
       // Use any type-scoped context defined on this value
-      const ctx = _getContextValue(typeContext, compactedType, '@context');
+      const ctx = _getContextValue(inputCtx, compactedType, '@context');
       if(!_isUndefined(ctx)) {
-        activeCtx = _processContext({
+        activeCtx = await _processContext({
           activeCtx,
           localCtx: ctx,
           options,
-          isTypeScopedContext: true
+          propagate: false
         });
       }
     }
@@ -185,18 +221,14 @@ api.compact = ({
     for(const expandedProperty of keys) {
       const expandedValue = element[expandedProperty];
 
-      // compact @id and @type(s)
-      if(expandedProperty === '@id' || expandedProperty === '@type') {
-        // if using a type-scoped context, resolve type values against previous
-        // context
-        const isType = expandedProperty === '@type';
-        const valueContext = isType ?
-          (activeCtx.previousContext || activeCtx) : activeCtx;
+      // compact @id
+      if(expandedProperty === '@id') {
         let compactedValue = _asArray(expandedValue).map(
           expandedIri => api.compactIri({
-            activeCtx: valueContext,
+            activeCtx,
             iri: expandedIri,
-            relativeTo: {vocab: isType}
+            relativeTo: {vocab: false},
+            base: options.base
           }));
         if(compactedValue.length === 1) {
           compactedValue = compactedValue[0];
@@ -204,8 +236,37 @@ api.compact = ({
 
         // use keyword alias and add value
         const alias = api.compactIri(
-          {activeCtx, iri: expandedProperty, relativeTo: {vocab: true}});
-        const isArray = _isArray(compactedValue) && expandedValue.length === 0;
+          {activeCtx, iri: '@id', relativeTo: {vocab: true}});
+
+        rval[alias] = compactedValue;
+        continue;
+      }
+
+      // compact @type(s)
+      if(expandedProperty === '@type') {
+        // resolve type values against previous context
+        let compactedValue = _asArray(expandedValue).map(
+          expandedIri => api.compactIri({
+            activeCtx: inputCtx,
+            iri: expandedIri,
+            relativeTo: {vocab: true}
+          }));
+        if(compactedValue.length === 1) {
+          compactedValue = compactedValue[0];
+        }
+
+        // use keyword alias and add value
+        const alias = api.compactIri(
+          {activeCtx, iri: '@type', relativeTo: {vocab: true}});
+        const container = _getContextValue(
+          activeCtx, alias, '@container') || [];
+
+        // treat as array for @type if @container includes @set
+        const typeAsSet =
+          container.includes('@set') &&
+          _processingMode(activeCtx, 1.1);
+        const isArray =
+          typeAsSet || (_isArray(compactedValue) && expandedValue.length === 0);
         _addValue(rval, alias, compactedValue, {propertyIsArray: isArray});
         continue;
       }
@@ -213,7 +274,7 @@ api.compact = ({
       // handle @reverse
       if(expandedProperty === '@reverse') {
         // recursively compact expanded value
-        const compactedValue = api.compact({
+        const compactedValue = await api.compact({
           activeCtx,
           activeProperty: '@reverse',
           element: expandedValue,
@@ -251,12 +312,13 @@ api.compact = ({
 
       if(expandedProperty === '@preserve') {
         // compact using activeProperty
-        const compactedValue = api.compact({
+        const compactedValue = await api.compact({
           activeCtx,
           activeProperty,
           element: expandedValue,
           options,
-          compactionMap});
+          compactionMap
+        });
 
         if(!(_isArray(compactedValue) && compactedValue.length === 0)) {
           _addValue(rval, expandedProperty, compactedValue);
@@ -283,8 +345,10 @@ api.compact = ({
         continue;
       }
 
-      // skip array processing for keywords that aren't @graph or @list
+      // skip array processing for keywords that aren't
+      // @graph, @list, or @included
       if(expandedProperty !== '@graph' && expandedProperty !== '@list' &&
+        expandedProperty !== '@included' &&
         _isKeyword(expandedProperty)) {
         // use keyword alias and add value as is
         const alias = api.compactIri({
@@ -366,7 +430,7 @@ api.compact = ({
         }
 
         // recursively compact expanded item
-        let compactedItem = api.compact({
+        let compactedItem = await api.compact({
           activeCtx,
           activeProperty: itemActiveProperty,
           element: (isList || isGraph) ? inner : expandedItem,
@@ -399,14 +463,12 @@ api.compact = ({
                 relativeTo: {vocab: true}
               })] = expandedItem['@index'];
             }
-          } else if(nestResult.hasOwnProperty(itemActiveProperty)) {
-            // can't use @list container for more than 1 list
-            throw new JsonLdError(
-              'JSON-LD compact error; property has a "@list" @container ' +
-              'rule but there is more than a single @list that matches ' +
-              'the compacted term in the document. Compaction might mix ' +
-              'unwanted items into the list.',
-              'jsonld.SyntaxError', {code: 'compaction to list of lists'});
+          } else {
+            _addValue(nestResult, itemActiveProperty, compactedItem, {
+              valueIsArray: true,
+              allowDuplicate: true
+            });
+            continue;
           }
         }
 
@@ -425,7 +487,8 @@ api.compact = ({
             // index on @id or @index or alias of @none
             const key = (container.includes('@id') ?
               expandedItem['@id'] : expandedItem['@index']) ||
-              api.compactIri({activeCtx, iri: '@none', vocab: true});
+              api.compactIri({activeCtx, iri: '@none',
+                relativeTo: {vocab: true}});
             // add compactedItem to map, using value of `@id` or a new blank
             // node identifier
 
@@ -438,6 +501,11 @@ api.compact = ({
             _isSimpleGraph(expandedItem)) {
             // container includes @graph but not @id or @index and value is a
             // simple graph object add compact value
+            // if compactedItem contains multiple values, it is wrapped in
+            // `@included`
+            if(_isArray(compactedItem) && compactedItem.length > 1) {
+              compactedItem = {'@included': compactedItem};
+            }
             _addValue(
               nestResult, itemActiveProperty, compactedItem, {
                 propertyIsArray:
@@ -502,16 +570,43 @@ api.compact = ({
             }
             key = expandedItem['@language'];
           } else if(container.includes('@index')) {
-            key = expandedItem['@index'];
+            const indexKey = _getContextValue(
+              activeCtx, itemActiveProperty, '@index') || '@index';
+            const containerKey = api.compactIri(
+              {activeCtx, iri: indexKey, relativeTo: {vocab: true}});
+            if(indexKey === '@index') {
+              key = expandedItem['@index'];
+              delete compactedItem[containerKey];
+            } else {
+              let others;
+              [key, ...others] = _asArray(compactedItem[indexKey] || []);
+              if(!_isString(key)) {
+                // Will use @none if it isn't a string.
+                key = null;
+              } else {
+                switch(others.length) {
+                  case 0:
+                    delete compactedItem[indexKey];
+                    break;
+                  case 1:
+                    compactedItem[indexKey] = others[0];
+                    break;
+                  default:
+                    compactedItem[indexKey] = others;
+                    break;
+                }
+              }
+            }
           } else if(container.includes('@id')) {
-            const idKey = api.compactIri({activeCtx, iri: '@id', vocab: true});
+            const idKey = api.compactIri({activeCtx, iri: '@id',
+              relativeTo: {vocab: true}});
             key = compactedItem[idKey];
             delete compactedItem[idKey];
           } else if(container.includes('@type')) {
             const typeKey = api.compactIri({
               activeCtx,
               iri: '@type',
-              vocab: true
+              relativeTo: {vocab: true}
             });
             let types;
             [key, ...types] = _asArray(compactedItem[typeKey] || []);
@@ -526,11 +621,25 @@ api.compact = ({
                 compactedItem[typeKey] = types;
                 break;
             }
+
+            // If compactedItem contains a single entry
+            // whose key maps to @id, recompact without @type
+            if(Object.keys(compactedItem).length === 1 &&
+              '@id' in expandedItem) {
+              compactedItem = await api.compact({
+                activeCtx,
+                activeProperty: itemActiveProperty,
+                element: {'@id': expandedItem['@id']},
+                options,
+                compactionMap
+              });
+            }
           }
 
           // if compacting this value which has no key, index on @none
           if(!key) {
-            key = api.compactIri({activeCtx, iri: '@none', vocab: true});
+            key = api.compactIri({activeCtx, iri: '@none',
+              relativeTo: {vocab: true}});
           }
           // add compact value to map object using key from expanded value
           // based on the container type
@@ -572,6 +681,7 @@ api.compact = ({
  * @param relativeTo options for how to compact IRIs:
  *          vocab: true to split after @vocab, false not to.
  * @param reverse true if a reverse property is being compacted, false if not.
+ * @param base the absolute URL to use for compacting document-relative IRIs.
  *
  * @return the compacted term, prefix, keyword alias, or the original IRI.
  */
@@ -580,7 +690,8 @@ api.compactIri = ({
   iri,
   value = null,
   relativeTo = {vocab: false},
-  reverse = false
+  reverse = false,
+  base = null
 }) => {
   // can't compact null
   if(iri === null) {
@@ -674,8 +785,12 @@ api.compactIri = ({
           let itemLanguage = '@none';
           let itemType = '@none';
           if(_isValue(item)) {
-            if('@language' in item) {
-              itemLanguage = item['@language'];
+            if('@direction' in item) {
+              const lang = (item['@language'] || '').toLowerCase();
+              const dir = item['@direction'];
+              itemLanguage = `${lang}_${dir}`;
+            } else if('@language' in item) {
+              itemLanguage = item['@language'].toLowerCase();
             } else if('@type' in item) {
               itemType = item['@type'];
             } else {
@@ -715,6 +830,12 @@ api.compactIri = ({
         if('@language' in value && !('@index' in value)) {
           containers.push('@language', '@language@set');
           typeOrLanguageValue = value['@language'];
+          const dir = value['@direction'];
+          if(dir) {
+            typeOrLanguageValue = `${typeOrLanguageValue}_${dir}`;
+          }
+        } else if('@direction' in value && !('@index' in value)) {
+          typeOrLanguageValue = `_${value['@direction']}`;
         } else if('@type' in value) {
           typeOrLanguage = '@type';
           typeOrLanguageValue = value['@type'];
@@ -806,9 +927,29 @@ api.compactIri = ({
     return choice;
   }
 
+  // If iri could be confused with a compact IRI using a term in this context,
+  // signal an error
+  for(const [term, td] of activeCtx.mappings) {
+    if(td && td._prefix && iri.startsWith(term + ':')) {
+      throw new JsonLdError(
+        `Absolute IRI "${iri}" confused with prefix "${term}".`,
+        'jsonld.SyntaxError',
+        {code: 'IRI confused with prefix', context: activeCtx});
+    }
+  }
+
   // compact IRI relative to base
   if(!relativeTo.vocab) {
-    return _removeBase(activeCtx['@base'], iri);
+    if('@base' in activeCtx) {
+      if(!activeCtx['@base']) {
+        // The None case preserves rval as potentially relative
+        return iri;
+      } else {
+        return _removeBase(_prependBase(base, activeCtx['@base']), iri);
+      }
+    } else {
+      return _removeBase(base, iri);
+    }
   }
 
   // return IRI as is
@@ -832,6 +973,7 @@ api.compactValue = ({activeCtx, activeProperty, value, options}) => {
     // get context rules
     const type = _getContextValue(activeCtx, activeProperty, '@type');
     const language = _getContextValue(activeCtx, activeProperty, '@language');
+    const direction = _getContextValue(activeCtx, activeProperty, '@direction');
     const container =
       _getContextValue(activeCtx, activeProperty, '@container') || [];
 
@@ -839,9 +981,19 @@ api.compactValue = ({activeCtx, activeProperty, value, options}) => {
     const preserveIndex = '@index' in value && !container.includes('@index');
 
     // if there's no @index to preserve ...
-    if(!preserveIndex) {
+    if(!preserveIndex && type !== '@none') {
       // matching @type or @language specified in context, compact value
-      if(value['@type'] === type || value['@language'] === language) {
+      if(value['@type'] === type) {
+        return value['@value'];
+      }
+      if('@language' in value && value['@language'] === language &&
+         '@direction' in value && value['@direction'] === direction) {
+        return value['@value'];
+      }
+      if('@language' in value && value['@language'] === language) {
+        return value['@value'];
+      }
+      if('@direction' in value && value['@direction'] === direction) {
         return value['@value'];
       }
     }
@@ -858,6 +1010,7 @@ api.compactValue = ({activeCtx, activeProperty, value, options}) => {
     const hasNullMapping = (activeCtx.mappings.has(activeProperty) &&
       activeCtx.mappings.get(activeProperty)['@language'] === null);
     if(isValueOnlyKey &&
+      type !== '@none' &&
       (!hasDefaultLanguage || !isValueString || hasNullMapping)) {
       return value['@value'];
     }
@@ -890,6 +1043,15 @@ api.compactValue = ({activeCtx, activeProperty, value, options}) => {
       })] = value['@language'];
     }
 
+    if('@direction' in value) {
+      // alias @direction
+      rval[api.compactIri({
+        activeCtx,
+        iri: '@direction',
+        relativeTo: {vocab: true}
+      })] = value['@direction'];
+    }
+
     // alias @value
     rval[api.compactIri({
       activeCtx,
@@ -904,8 +1066,11 @@ api.compactValue = ({activeCtx, activeProperty, value, options}) => {
   const expandedProperty = _expandIri(activeCtx, activeProperty, {vocab: true},
     options);
   const type = _getContextValue(activeCtx, activeProperty, '@type');
-  const compacted = api.compactIri(
-    {activeCtx, iri: value['@id'], relativeTo: {vocab: type === '@vocab'}});
+  const compacted = api.compactIri({
+    activeCtx,
+    iri: value['@id'],
+    relativeTo: {vocab: type === '@vocab'},
+    base: options.base});
 
   // compact to scalar
   if(type === '@id' || type === '@vocab' || expandedProperty === '@graph') {
@@ -921,95 +1086,6 @@ api.compactValue = ({activeCtx, activeProperty, value, options}) => {
   };
 };
 
-/**
- * Removes the @preserve keywords as the last step of the compaction
- * algorithm when it is running on framed output.
- *
- * @param ctx the active context used to compact the input.
- * @param input the framed, compacted output.
- * @param options the compaction options used.
- *
- * @return the resulting output.
- */
-api.removePreserve = (ctx, input, options) => {
-  // recurse through arrays
-  if(_isArray(input)) {
-    const output = [];
-    for(let i = 0; i < input.length; ++i) {
-      const result = api.removePreserve(ctx, input[i], options);
-      // drop nulls from arrays
-      if(result !== null) {
-        output.push(result);
-      }
-    }
-    input = output;
-  } else if(_isObject(input)) {
-    // remove @preserve
-    if('@preserve' in input) {
-      if(input['@preserve'] === '@null') {
-        return null;
-      }
-      return input['@preserve'];
-    }
-
-    // skip @values
-    if(_isValue(input)) {
-      return input;
-    }
-
-    // recurse through @lists
-    if(_isList(input)) {
-      input['@list'] = api.removePreserve(ctx, input['@list'], options);
-      return input;
-    }
-
-    // handle in-memory linked nodes
-    const idAlias = api.compactIri({
-      activeCtx: ctx,
-      iri: '@id',
-      relativeTo: {vocab: true}
-    });
-    if(input.hasOwnProperty(idAlias)) {
-      const id = input[idAlias];
-      if(options.link.hasOwnProperty(id)) {
-        const idx = options.link[id].indexOf(input);
-        if(idx !== -1) {
-          // already visited
-          return options.link[id][idx];
-        }
-        // prevent circular visitation
-        options.link[id].push(input);
-      } else {
-        // prevent circular visitation
-        options.link[id] = [input];
-      }
-    }
-
-    // recurse through properties
-    const graphAlias = api.compactIri({
-      activeCtx: ctx,
-      iri: '@graph',
-      relativeTo: {vocab: true}
-    });
-    for(const prop in input) {
-      // potentially remove the id, if it is an unreference bnode
-      if(prop === idAlias && options.bnodesToClear.includes(input[prop])) {
-        delete input[idAlias];
-        continue;
-      }
-
-      let result = api.removePreserve(ctx, input[prop], options);
-      const container = _getContextValue(ctx, prop, '@container') || [];
-      if(options.compactArrays && _isArray(result) && result.length === 1 &&
-        container.length === 0 && prop !== graphAlias) {
-        result = result[0];
-      }
-      input[prop] = result;
-    }
-  }
-  return input;
-};
-
 /**
  * Picks the preferred compaction term from the given inverse context entry.
  *
@@ -1033,7 +1109,7 @@ function _selectTerm(
 
   // determine prefs for @id based on whether or not value compacts to a term
   if((typeOrLanguageValue === '@id' || typeOrLanguageValue === '@reverse') &&
-    _isSubjectReference(value)) {
+    _isObject(value) && '@id' in value) {
     // prefer @reverse first
     if(typeOrLanguageValue === '@reverse') {
       prefs.push('@reverse');
@@ -1052,21 +1128,26 @@ function _selectTerm(
     }
   } else {
     prefs.push(typeOrLanguageValue);
+
+    // consider direction only
+    const langDir = prefs.find(el => el.includes('_'));
+    if(langDir) {
+      // consider _dir portion
+      prefs.push(langDir.replace(/^[^_]+_/, '_'));
+    }
   }
   prefs.push('@none');
 
   const containerMap = activeCtx.inverse[iri];
-  for(let ci = 0; ci < containers.length; ++ci) {
+  for(const container of containers) {
     // if container not available in the map, continue
-    const container = containers[ci];
     if(!(container in containerMap)) {
       continue;
     }
 
     const typeOrLanguageValueMap = containerMap[container][typeOrLanguage];
-    for(let pi = 0; pi < prefs.length; ++pi) {
+    for(const pref of prefs) {
       // if type/language option not available in the map, continue
-      const pref = prefs[pi];
       if(!(pref in typeOrLanguageValueMap)) {
         continue;
       }
