@@ -16,8 +16,7 @@ const {
 const {
   isAbsolute: _isAbsoluteIri,
   isRelative: _isRelativeIri,
-  prependBase,
-  parse: parseUrl
+  prependBase
 } = require('./url');
 
 const {
@@ -922,7 +921,7 @@ api.createTermDefinition = ({
       const protectedMode = (options && options.protectedMode) || 'error';
       if(protectedMode === 'error') {
         throw new JsonLdError(
-          'Invalid JSON-LD syntax; tried to redefine a protected term.',
+          `Invalid JSON-LD syntax; tried to redefine "${term}" which is a protected term.`,
           'jsonld.SyntaxError',
           {code: 'protected term redefinition', context: localCtx, term});
       } else if(protectedMode === 'warn') {
