@@ -4,9 +4,10 @@
 'use strict';
 
 const {parseLinkHeader, buildHeaders} = require('../util');
-const {LINK_HEADER_REL} = require('../constants');
+const {LINK_HEADER_CONTEXT} = require('../constants');
 const JsonLdError = require('../JsonLdError');
 const RequestQueue = require('../RequestQueue');
+const {prependBase} = require('../url');
 
 /**
  * Creates a built-in node document loader.
@@ -32,11 +33,9 @@ module.exports = ({
   headers = {}
 } = {strictSSL: true, maxRedirects: -1, headers: {}}) => {
   headers = buildHeaders(headers);
-  // TODO: use `r2`
+  // TODO: use `axios`
   request = request || require('request');
   const http = require('http');
-  // TODO: disable cache until HTTP caching implemented
-  //const cache = new DocumentCache();
 
   const queue = new RequestQueue();
   return queue.wrapLoader(function(url) {
@@ -63,6 +62,7 @@ module.exports = ({
     }
 
     let result;
+    let alternate = null;
     try {
       result = await _request(request, {
         url,
@@ -85,7 +85,7 @@ module.exports = ({
     const statusText = http.STATUS_CODES[res.statusCode];
     if(res.statusCode >= 400) {
       throw new JsonLdError(
-        'URL could not be dereferenced: ' + statusText,
+        `URL "${url}" could not be dereferenced: ${statusText}`,
         'jsonld.InvalidUrl', {
           code: 'loading document failed',
           url,
@@ -97,21 +97,32 @@ module.exports = ({
     if(res.headers.link &&
       res.headers['content-type'] !== 'application/ld+json') {
       // only 1 related link header permitted
-      const linkHeader = parseLinkHeader(res.headers.link)[LINK_HEADER_REL];
-      if(Array.isArray(linkHeader)) {
+      const linkHeaders = parseLinkHeader(res.headers.link);
+      const linkedContext = linkHeaders[LINK_HEADER_CONTEXT];
+      if(Array.isArray(linkedContext)) {
         throw new JsonLdError(
           'URL could not be dereferenced, it has more than one associated ' +
           'HTTP Link Header.',
           'jsonld.InvalidUrl',
           {code: 'multiple context link headers', url});
       }
-      if(linkHeader) {
-        doc.contextUrl = linkHeader.target;
+      if(linkedContext) {
+        doc.contextUrl = linkedContext.target;
+      }
+
+      // "alternate" link header is a redirect
+      alternate = linkHeaders['alternate'];
+      if(alternate &&
+        alternate.type == 'application/ld+json' &&
+        !(res.headers['content-type'] || '')
+          .match(/^application\/(\w*\+)?json$/)) {
+        res.headers.location = prependBase(url, alternate.target);
       }
     }
 
     // handle redirect
-    if(res.statusCode >= 300 && res.statusCode < 400 && res.headers.location) {
+    if((alternate ||
+      res.statusCode >= 300 && res.statusCode < 400) && res.headers.location) {
       if(redirects.length === maxRedirects) {
         throw new JsonLdError(
           'URL could not be dereferenced; there were too many redirects.',
