@@ -4,9 +4,10 @@
 'use strict';
 
 const {parseLinkHeader, buildHeaders} = require('../util');
-const {LINK_HEADER_REL} = require('../constants');
+const {LINK_HEADER_CONTEXT} = require('../constants');
 const JsonLdError = require('../JsonLdError');
 const RequestQueue = require('../RequestQueue');
+const {prependBase} = require('../url');
 
 const REGEX_LINK_HEADER = /(^|(\r\n))link:/i;
 
@@ -64,7 +65,8 @@ module.exports = ({
         });
     }
 
-    const doc = {contextUrl: null, documentUrl: url, document: req.response};
+    let doc = {contextUrl: null, documentUrl: url, document: req.response};
+    let alternate = null;
 
     // handle Link Header (avoid unsafe header warning by existence testing)
     const contentType = req.getResponseHeader('Content-Type');
@@ -74,16 +76,25 @@ module.exports = ({
     }
     if(linkHeader && contentType !== 'application/ld+json') {
       // only 1 related link header permitted
-      linkHeader = parseLinkHeader(linkHeader)[LINK_HEADER_REL];
-      if(Array.isArray(linkHeader)) {
+      const linkHeaders = parseLinkHeader(linkHeader);
+      const linkedContext = linkHeaders[LINK_HEADER_CONTEXT];
+      if(Array.isArray(linkedContext)) {
         throw new JsonLdError(
           'URL could not be dereferenced, it has more than one ' +
           'associated HTTP Link Header.',
           'jsonld.InvalidUrl',
           {code: 'multiple context link headers', url});
       }
-      if(linkHeader) {
-        doc.contextUrl = linkHeader.target;
+      if(linkedContext) {
+        doc.contextUrl = linkedContext.target;
+      }
+
+      // "alternate" link header is a redirect
+      alternate = linkHeaders['alternate'];
+      if(alternate &&
+        alternate.type == 'application/ld+json' &&
+        !(contentType || '').match(/^application\/(\w*\+)?json$/)) {
+        doc = await loader(prependBase(url, alternate.target));
       }
     }
 
