@@ -16,14 +16,16 @@ const {
 const {
   isList: _isList,
   isValue: _isValue,
-  isGraph: _isGraph
+  isGraph: _isGraph,
+  isSubject: _isSubject
 } = require('./graphTypes');
 
 const {
   expandIri: _expandIri,
   getContextValue: _getContextValue,
   isKeyword: _isKeyword,
-  process: _processContext
+  process: _processContext,
+  processingMode: _processingMode
 } = require('./context');
 
 const {
@@ -39,6 +41,7 @@ const {
 
 const api = {};
 module.exports = api;
+const REGEX_BCP47 = /^[a-zA-Z]{1,8}(-[a-zA-Z0-9]{1,8})*$/;
 
 /**
  * Recursively expands an element using the given context. Any context in
@@ -62,7 +65,7 @@ module.exports = api;
  *
  * @return a Promise that resolves to the expanded value.
  */
-api.expand = ({
+api.expand = async ({
   activeCtx,
   activeProperty = null,
   element,
@@ -87,8 +90,7 @@ api.expand = ({
     if(!insideList && (activeProperty === null ||
       _expandIri(activeCtx, activeProperty, {vocab: true},
         options) === '@graph')) {
-      // TODO: use `await` to support async
-      const mapped = expansionMap({
+      const mapped = await expansionMap({
         unmappedValue: element,
         activeCtx,
         activeProperty,
@@ -113,7 +115,7 @@ api.expand = ({
     insideList = insideList || container.includes('@list');
     for(let i = 0; i < element.length; ++i) {
       // expand element
-      let e = api.expand({
+      let e = await api.expand({
         activeCtx,
         activeProperty,
         element: element[i],
@@ -122,16 +124,12 @@ api.expand = ({
         insideIndex,
         typeScopedContext
       });
-      if(insideList && (_isArray(e) || _isList(e))) {
-        // lists of lists are illegal
-        throw new JsonLdError(
-          'Invalid JSON-LD syntax; lists of lists are not permitted.',
-          'jsonld.SyntaxError', {code: 'list of lists'});
+      if(insideList && _isArray(e)) {
+        e = {'@list': e};
       }
 
       if(e === null) {
-        // TODO: add `await` for async support
-        e = expansionMap({
+        e = await expansionMap({
           unmappedValue: element[i],
           activeCtx,
           activeProperty,
@@ -161,6 +159,10 @@ api.expand = ({
   const expandedActiveProperty = _expandIri(
     activeCtx, activeProperty, {vocab: true}, options);
 
+  // Get any property-scoped context for activeProperty
+  const propertyScopedCtx =
+    _getContextValue(activeCtx, activeProperty, '@context');
+
   // second, determine if any type-scoped context should be reverted; it
   // should only be reverted when the following are all true:
   // 1. `element` is not a value or subject reference
@@ -190,34 +192,51 @@ api.expand = ({
 
   if(mustRevert) {
     // revert type scoped context
-    activeCtx = activeCtx.revertTypeScopedContext();
+    activeCtx = activeCtx.revertToPreviousContext();
+  }
+
+  // apply property-scoped context after reverting term-scoped context
+  if(!_isUndefined(propertyScopedCtx)) {
+    activeCtx = await _processContext({
+      activeCtx,
+      localCtx: propertyScopedCtx,
+      propagate: true,
+      overrideProtected: true,
+      options
+    });
   }
 
   // if element has a context, process it
   if('@context' in element) {
-    activeCtx = _processContext(
+    activeCtx = await _processContext(
       {activeCtx, localCtx: element['@context'], options});
   }
 
+  // set the type-scoped context to the context on input, for use later
+  typeScopedContext = activeCtx;
+
+  // Remember the first key found expanding to @type
+  let typeKey = null;
+
   // look for scoped contexts on `@type`
   for(const key of keys) {
     const expandedProperty = _expandIri(activeCtx, key, {vocab: true}, options);
     if(expandedProperty === '@type') {
       // set scoped contexts from @type
       // avoid sorting if possible
+      typeKey = typeKey || key;
       const value = element[key];
       const types =
         Array.isArray(value) ?
           (value.length > 1 ? value.slice().sort() : value) : [value];
       for(const type of types) {
-        const ctx = _getContextValue(
-          activeCtx.previousContext || activeCtx, type, '@context');
+        const ctx = _getContextValue(typeScopedContext, type, '@context');
         if(!_isUndefined(ctx)) {
-          activeCtx = _processContext({
+          activeCtx = await _processContext({
             activeCtx,
             localCtx: ctx,
             options,
-            isTypeScopedContext: true
+            propagate: false
           });
         }
       }
@@ -226,7 +245,7 @@ api.expand = ({
 
   // process each key and value in element, ignoring @nest content
   let rval = {};
-  _expandObject({
+  await _expandObject({
     activeCtx,
     activeProperty,
     expandedActiveProperty,
@@ -234,6 +253,7 @@ api.expand = ({
     expandedParent: rval,
     options,
     insideList,
+    typeKey,
     typeScopedContext,
     expansionMap});
 
@@ -243,10 +263,10 @@ api.expand = ({
 
   if('@value' in rval) {
     // @value must only have @language or @type
-    if('@type' in rval && '@language' in rval) {
+    if('@type' in rval && ('@language' in rval || '@direction' in rval)) {
       throw new JsonLdError(
         'Invalid JSON-LD syntax; an element containing "@value" may not ' +
-        'contain both "@type" and "@language".',
+        'contain both "@type" and either "@language" or "@direction".',
         'jsonld.SyntaxError', {code: 'invalid value object', element: rval});
     }
     let validCount = count - 1;
@@ -259,20 +279,25 @@ api.expand = ({
     if('@language' in rval) {
       validCount -= 1;
     }
+    if('@direction' in rval) {
+      validCount -= 1;
+    }
     if(validCount !== 0) {
       throw new JsonLdError(
         'Invalid JSON-LD syntax; an element containing "@value" may only ' +
-        'have an "@index" property and at most one other property ' +
-        'which can be "@type" or "@language".',
+        'have an "@index" property and either "@type" ' +
+        'or either or both "@language" or "@direction".',
         'jsonld.SyntaxError', {code: 'invalid value object', element: rval});
     }
     const values = rval['@value'] === null ? [] : _asArray(rval['@value']);
     const types = _getValues(rval, '@type');
 
     // drop null @values unless custom mapped
-    if(values.length === 0) {
-      // TODO: use `await` to support async
-      const mapped = expansionMap({
+    if(_processingMode(activeCtx, 1.1) && types.includes('@json') &&
+      types.length === 1) {
+      // Any value of @value is okay if @type: @json
+    } else if(values.length === 0) {
+      const mapped = await expansionMap({
         unmappedValue: rval,
         activeCtx,
         activeProperty,
@@ -320,8 +345,7 @@ api.expand = ({
     }
   } else if(count === 1 && '@language' in rval) {
     // drop objects with only @language unless custom mapped
-    // TODO: use `await` to support async
-    const mapped = expansionMap(rval, {
+    const mapped = await expansionMap(rval, {
       unmappedValue: rval,
       activeCtx,
       activeProperty,
@@ -344,8 +368,7 @@ api.expand = ({
     // drop empty object, top-level @value/@list, or object with only @id
     if(count === 0 || '@value' in rval || '@list' in rval ||
       (count === 1 && '@id' in rval)) {
-      // TODO: use `await` to support async
-      const mapped = expansionMap({
+      const mapped = await expansionMap({
         unmappedValue: rval,
         activeCtx,
         activeProperty,
@@ -374,12 +397,14 @@ api.expand = ({
  * @param expandedParent the expanded result into which to add values.
  * @param options the expansion options.
  * @param insideList true if the element is a list, false if not.
+ * @param typeKey first key found expanding to @type.
+ * @param typeScopedContext the context before reverting.
  * @param expansionMap(info) a function that can be used to custom map
  *          unmappable values (or to throw an error when they are detected);
  *          if this function returns `undefined` then the default behavior
  *          will be used.
  */
-function _expandObject({
+async function _expandObject({
   activeCtx,
   activeProperty,
   expandedActiveProperty,
@@ -387,10 +412,20 @@ function _expandObject({
   expandedParent,
   options = {},
   insideList,
+  typeKey,
+  typeScopedContext,
   expansionMap
 }) {
   const keys = Object.keys(element).sort();
   const nests = [];
+  let unexpandedValue;
+
+  // Figure out if this is the type for a JSON literal
+  const isJsonType = element[typeKey] &&
+    _expandIri(activeCtx,
+      (_isArray(element[typeKey]) ? element[typeKey][0] : element[typeKey]),
+      {vocab: true}, options) === '@json';
+
   for(const key of keys) {
     let value = element[key];
     let expandedValue;
@@ -429,7 +464,9 @@ function _expandObject({
           'property.', 'jsonld.SyntaxError',
           {code: 'invalid reverse property map', value});
       }
-      if(expandedProperty in expandedParent) {
+      if(expandedProperty in expandedParent &&
+         expandedProperty !== '@included' &&
+         expandedProperty !== '@type') {
         throw new JsonLdError(
           'Invalid JSON-LD syntax; colliding keywords detected.',
           'jsonld.SyntaxError',
@@ -477,17 +514,52 @@ function _expandObject({
     }
 
     if(expandedProperty === '@type') {
-      _validateTypeValue(value);
+      // if framing, can be a default object, but need to expand
+      // key to determine that
+      if(_isObject(value)) {
+        value = Object.fromEntries(Object.entries(value).map(([k, v]) => [
+          _expandIri(typeScopedContext, k, {vocab: true}),
+          _asArray(v).map(vv =>
+            _expandIri(typeScopedContext, vv, {base: true, vocab: true})
+          )
+        ]));
+      }
+      _validateTypeValue(value, options.isFrame);
       _addValue(
         expandedParent, '@type',
         _asArray(value).map(v =>
           _isString(v) ?
-            _expandIri(activeCtx.previousContext || activeCtx, v,
+            _expandIri(typeScopedContext, v,
               {base: true, vocab: true}, options) : v),
         {propertyIsArray: options.isFrame});
       continue;
     }
 
+    // Included blocks are treated as an array of separate object nodes sharing
+    // the same referencing active_property.
+    // For 1.0, it is skipped as are other unknown keywords
+    if(expandedProperty === '@included' && _processingMode(activeCtx, 1.1)) {
+      const includedResult = _asArray(await api.expand({
+        activeCtx,
+        activeProperty,
+        element: value,
+        options,
+        expansionMap
+      }));
+
+      // Expanded values must be node objects
+      if(!includedResult.every(v => _isSubject(v))) {
+        throw new JsonLdError(
+          'Invalid JSON-LD syntax; ' +
+          'values of @included must expand to node objects.',
+          'jsonld.SyntaxError', {code: 'invalid @included value', value});
+      }
+
+      _addValue(
+        expandedParent, '@included', includedResult, {propertyIsArray: true});
+      continue;
+    }
+
     // @graph must be an array or an object
     if(expandedProperty === '@graph' &&
       !(_isObject(value) || _isArray(value))) {
@@ -497,22 +569,22 @@ function _expandObject({
         'jsonld.SyntaxError', {code: 'invalid @graph value', value});
     }
 
-    // @value must not be an object or an array (unless framing)
     if(expandedProperty === '@value') {
-      if((_isObject(value) || _isArray(value)) && !options.isFrame) {
-        throw new JsonLdError(
-          'Invalid JSON-LD syntax; "@value" value must not be an ' +
-          'object or an array.',
-          'jsonld.SyntaxError',
-          {code: 'invalid value object value', value});
+      // capture value for later
+      // "colliding keywords" check prevents this from being set twice
+      unexpandedValue = value;
+      if(isJsonType && _processingMode(activeCtx, 1.1)) {
+        // no coercion to array, and retain all values
+        expandedParent['@value'] = value;
+      } else {
+        _addValue(
+          expandedParent, '@value', value, {propertyIsArray: options.isFrame});
       }
-
-      _addValue(
-        expandedParent, '@value', value, {propertyIsArray: options.isFrame});
       continue;
     }
 
     // @language must be a string
+    // it should match BCP47
     if(expandedProperty === '@language') {
       if(value === null) {
         // drop null @language values, they expand as if they didn't exist
@@ -527,11 +599,45 @@ function _expandObject({
       // ensure language value is lowercase
       value = _asArray(value).map(v => _isString(v) ? v.toLowerCase() : v);
 
+      // ensure language tag matches BCP47
+      for(const lang of value) {
+        if(_isString(lang) && !lang.match(REGEX_BCP47)) {
+          console.warn(`@language must be valid BCP47: ${lang}`);
+        }
+      }
+
       _addValue(
         expandedParent, '@language', value, {propertyIsArray: options.isFrame});
       continue;
     }
 
+    // @direction must be "ltr" or "rtl"
+    if(expandedProperty === '@direction') {
+      if(!_isString(value) && !options.isFrame) {
+        throw new JsonLdError(
+          'Invalid JSON-LD syntax; "@direction" value must be a string.',
+          'jsonld.SyntaxError',
+          {code: 'invalid base direction', value});
+      }
+
+      value = _asArray(value);
+
+      // ensure direction is "ltr" or "rtl"
+      for(const dir of value) {
+        if(_isString(dir) && dir !== 'ltr' && dir !== 'rtl') {
+          throw new JsonLdError(
+            'Invalid JSON-LD syntax; "@direction" must be "ltr" or "rtl".',
+            'jsonld.SyntaxError',
+            {code: 'invalid base direction', value});
+        }
+      }
+
+      _addValue(
+        expandedParent, '@direction', value,
+        {propertyIsArray: options.isFrame});
+      continue;
+    }
+
     // @index must be a string
     if(expandedProperty === '@index') {
       if(!_isString(value)) {
@@ -552,7 +658,7 @@ function _expandObject({
           'jsonld.SyntaxError', {code: 'invalid @reverse value', value});
       }
 
-      expandedValue = api.expand({
+      expandedValue = await api.expand({
         activeCtx,
         activeProperty:
         '@reverse',
@@ -606,11 +712,11 @@ function _expandObject({
     let termCtx = activeCtx;
     const ctx = _getContextValue(activeCtx, key, '@context');
     if(!_isUndefined(ctx)) {
-      // Note: spec's `from term` var is named `isPropertyTermScopedContext`
-      termCtx = _processContext({
+      termCtx = await _processContext({
         activeCtx,
         localCtx: ctx,
-        isPropertyTermScopedContext: true,
+        propagate: true,
+        overrideProtected: true,
         options
       });
     }
@@ -618,24 +724,30 @@ function _expandObject({
     const container = _getContextValue(termCtx, key, '@container') || [];
 
     if(container.includes('@language') && _isObject(value)) {
+      const direction = _getContextValue(termCtx, key, '@direction');
       // handle language map container (skip if value is not an object)
-      expandedValue = _expandLanguageMap(termCtx, value, options);
+      expandedValue = _expandLanguageMap(termCtx, value, direction, options);
     } else if(container.includes('@index') && _isObject(value)) {
       // handle index container (skip if value is not an object)
       const asGraph = container.includes('@graph');
-      expandedValue = _expandIndexMap({
+      const indexKey = _getContextValue(termCtx, key, '@index') || '@index';
+      const propertyIndex = indexKey !== '@index' &&
+        _expandIri(activeCtx, indexKey, {vocab: true}, options);
+
+      expandedValue = await _expandIndexMap({
         activeCtx: termCtx,
         options,
         activeProperty: key,
         value,
         expansionMap,
         asGraph,
-        indexKey: '@index'
+        indexKey,
+        propertyIndex
       });
     } else if(container.includes('@id') && _isObject(value)) {
       // handle id container (skip if value is not an object)
       const asGraph = container.includes('@graph');
-      expandedValue = _expandIndexMap({
+      expandedValue = await _expandIndexMap({
         activeCtx: termCtx,
         options,
         activeProperty: key,
@@ -646,9 +758,9 @@ function _expandObject({
       });
     } else if(container.includes('@type') && _isObject(value)) {
       // handle type container (skip if value is not an object)
-      expandedValue = _expandIndexMap({
+      expandedValue = await _expandIndexMap({
         // since container is `@type`, revert type scoped context when expanding
-        activeCtx: termCtx.revertTypeScopedContext(),
+        activeCtx: termCtx.revertToPreviousContext(),
         options,
         activeProperty: key,
         value,
@@ -664,7 +776,7 @@ function _expandObject({
         if(isList && expandedActiveProperty === '@graph') {
           nextActiveProperty = null;
         }
-        expandedValue = api.expand({
+        expandedValue = await api.expand({
           activeCtx: termCtx,
           activeProperty: nextActiveProperty,
           element: value,
@@ -672,14 +784,15 @@ function _expandObject({
           insideList: isList,
           expansionMap
         });
-        if(isList && _isList(expandedValue)) {
-          throw new JsonLdError(
-            'Invalid JSON-LD syntax; lists of lists are not permitted.',
-            'jsonld.SyntaxError', {code: 'list of lists'});
-        }
+      } else if(
+        _getContextValue(activeCtx, key, '@type') === '@json') {
+        expandedValue = {
+          '@type': '@json',
+          '@value': value
+        };
       } else {
         // recursively expand value with key as new active property
-        expandedValue = api.expand({
+        expandedValue = await api.expand({
           activeCtx: termCtx,
           activeProperty: key,
           element: value,
@@ -723,7 +836,7 @@ function _expandObject({
       !container.some(key => key === '@id' || key === '@index')) {
       // ensure expanded values are arrays
       expandedValue = _asArray(expandedValue)
-        .map(v => _isGraph(v) ? v : {'@graph': _asArray(v)});
+        .map(v => ({'@graph': _asArray(v)}));
     }
 
     // FIXME: can this be merged with code above to simplify?
@@ -746,15 +859,28 @@ function _expandObject({
     }
 
     // add value for property
-    // use an array except for certain keywords
-    const useArray =
-      !['@index', '@id', '@type', '@value', '@language']
-        .includes(expandedProperty);
+    // special keywords handled above
     _addValue(expandedParent, expandedProperty, expandedValue, {
-      propertyIsArray: useArray
+      propertyIsArray: true
     });
   }
 
+  // @value must not be an object or an array (unless framing) or if @type is
+  // @json
+  if('@value' in expandedParent) {
+    if(expandedParent['@type'] === '@json' && _processingMode(activeCtx, 1.1)) {
+      // allow any value, to be verified when the object is fully expanded and
+      // the @type is @json.
+    } else if((_isObject(unexpandedValue) || _isArray(unexpandedValue)) &&
+      !options.isFrame) {
+      throw new JsonLdError(
+        'Invalid JSON-LD syntax; "@value" value must not be an ' +
+        'object or an array.',
+        'jsonld.SyntaxError',
+        {code: 'invalid value object value', value: unexpandedValue});
+    }
+  }
+
   // expand each nested key
   for(const key of nests) {
     const nestedValues = _isArray(element[key]) ? element[key] : [element[key]];
@@ -766,7 +892,7 @@ function _expandObject({
           'jsonld.SyntaxError',
           {code: 'invalid @nest value', value: nv});
       }
-      _expandObject({
+      await _expandObject({
         activeCtx,
         activeProperty,
         expandedActiveProperty,
@@ -774,6 +900,8 @@ function _expandObject({
         expandedParent,
         options,
         insideList,
+        typeScopedContext,
+        typeKey,
         expansionMap});
     }
   }
@@ -826,7 +954,7 @@ function _expandValue({activeCtx, activeProperty, value, options}) {
 
   const rval = {};
 
-  if(type && !['@id', '@vocab'].includes(type)) {
+  if(type && !['@id', '@vocab', '@none'].includes(type)) {
     // other type
     rval['@type'] = type;
   } else if(_isString(value)) {
@@ -835,6 +963,10 @@ function _expandValue({activeCtx, activeProperty, value, options}) {
     if(language !== null) {
       rval['@language'] = language;
     }
+    const direction = _getContextValue(activeCtx, activeProperty, '@direction');
+    if(direction !== null) {
+      rval['@direction'] = direction;
+    }
   }
   // do conversion of values that aren't basic JSON types to strings
   if(!['boolean', 'number', 'string'].includes(typeof value)) {
@@ -850,11 +982,12 @@ function _expandValue({activeCtx, activeProperty, value, options}) {
  *
  * @param activeCtx the active context to use.
  * @param languageMap the language map to expand.
+ * @param direction the direction to apply to values.
  * @param {Object} [options] - processing options.
  *
  * @return the expanded language map.
  */
-function _expandLanguageMap(activeCtx, languageMap, options) {
+function _expandLanguageMap(activeCtx, languageMap, direction, options) {
   const rval = [];
   const keys = Object.keys(languageMap).sort();
   for(const key of keys) {
@@ -878,15 +1011,18 @@ function _expandLanguageMap(activeCtx, languageMap, options) {
       if(expandedKey !== '@none') {
         val['@language'] = key.toLowerCase();
       }
+      if(direction) {
+        val['@direction'] = direction;
+      }
       rval.push(val);
     }
   }
   return rval;
 }
 
-function _expandIndexMap(
+async function _expandIndexMap(
   {activeCtx, options, activeProperty, value, expansionMap, asGraph,
-    indexKey}) {
+    indexKey, propertyIndex}) {
   const rval = [];
   const keys = Object.keys(value).sort();
   const isTypeIndex = indexKey === '@type';
@@ -895,10 +1031,10 @@ function _expandIndexMap(
     if(isTypeIndex) {
       const ctx = _getContextValue(activeCtx, key, '@context');
       if(!_isUndefined(ctx)) {
-        activeCtx = _processContext({
+        activeCtx = await _processContext({
           activeCtx,
           localCtx: ctx,
-          isTypeScopedContext: true,
+          propagate: false,
           options
         });
       }
@@ -909,16 +1045,7 @@ function _expandIndexMap(
       val = [val];
     }
 
-    // expand for @type, but also for @none
-    const expandedKey = _expandIri(activeCtx, key, {vocab: true}, options);
-    if(indexKey === '@id') {
-      // expand document relative
-      key = _expandIri(activeCtx, key, {base: true}, options);
-    } else if(isTypeIndex) {
-      key = expandedKey;
-    }
-
-    val = api.expand({
+    val = await api.expand({
       activeCtx,
       activeProperty,
       element: val,
@@ -927,6 +1054,27 @@ function _expandIndexMap(
       insideIndex: true,
       expansionMap
     });
+
+    // expand for @type, but also for @none
+    let expandedKey;
+    if(propertyIndex) {
+      if(key === '@none') {
+        expandedKey = '@none';
+      } else {
+        expandedKey = _expandValue(
+          {activeCtx, activeProperty: indexKey, value: key, options});
+      }
+    } else {
+      expandedKey = _expandIri(activeCtx, key, {vocab: true}, options);
+    }
+
+    if(indexKey === '@id') {
+      // expand document relative
+      key = _expandIri(activeCtx, key, {base: true}, options);
+    } else if(isTypeIndex) {
+      key = expandedKey;
+    }
+
     for(let item of val) {
       // If this is also a @graph container, turn items into graphs
       if(asGraph && !_isGraph(item)) {
@@ -940,6 +1088,23 @@ function _expandIndexMap(
         } else {
           item['@type'] = [key];
         }
+      } else if(_isValue(item) &&
+        !['@language', '@type', '@index'].includes(indexKey)) {
+        throw new JsonLdError(
+          'Invalid JSON-LD syntax; Attempt to add illegal key to value ' +
+          `object: "${indexKey}".`,
+          'jsonld.SyntaxError',
+          {code: 'invalid value object', value: item});
+      } else if(propertyIndex) {
+        // index is a property to be expanded, and values interpreted for that
+        // property
+        if(expandedKey !== '@none') {
+          // expand key as a value
+          _addValue(item, propertyIndex, expandedKey, {
+            propertyIsArray: true,
+            prependValue: true
+          });
+        }
       } else if(expandedKey !== '@none' && !(indexKey in item)) {
         item[indexKey] = key;
       }
