@@ -7,6 +7,7 @@ const {isKeyword} = require('./context');
 const graphTypes = require('./graphTypes');
 const types = require('./types');
 const util = require('./util');
+const url = require('./url');
 const JsonLdError = require('./JsonLdError');
 const {
   createNodeMap: _createNodeMap,
@@ -29,9 +30,9 @@ api.frameMergedOrDefault = (input, frame, options) => {
   // create framing state
   const state = {
     options,
+    embedded: false,
     graph: '@default',
     graphMap: {'@default': {}},
-    graphStack: [],
     subjectStack: [],
     link: {},
     bnodeMap: {}
@@ -58,7 +59,9 @@ api.frameMergedOrDefault = (input, frame, options) => {
       Object.keys(state.bnodeMap).filter(id => state.bnodeMap[id].length === 1);
   }
 
-  return framed;
+  // remove @preserve from results
+  options.link = {};
+  return _cleanupPreserve(framed, options);
 };
 
 /**
@@ -83,6 +86,12 @@ api.frame = (state, subjects, frame, parent, property = null) => {
     requireAll: _getFrameFlag(frame, options, 'requireAll')
   };
 
+  // get link for current graph
+  if(!state.link.hasOwnProperty(state.graph)) {
+    state.link[state.graph] = {};
+  }
+  const link = state.link[state.graph];
+
   // filter out subjects that match the frame
   const matches = _filterSubjects(state, subjects, frame, flags);
 
@@ -91,16 +100,6 @@ api.frame = (state, subjects, frame, parent, property = null) => {
   for(const id of ids) {
     const subject = matches[id];
 
-    if(flags.embed === '@link' && id in state.link) {
-      // TODO: may want to also match an existing linked subject against
-      // the current frame ... so different frames could produce different
-      // subjects that are only shared in-memory when the frames are the same
-
-      // add existing linked subject
-      _addFrameOutput(parent, property, state.link[id]);
-      continue;
-    }
-
     /* Note: In order to treat each top-level match as a compartmentalized
     result, clear the unique embedded subjects map when the property is null,
     which only occurs at the top-level. */
@@ -110,20 +109,51 @@ api.frame = (state, subjects, frame, parent, property = null) => {
       state.uniqueEmbeds[state.graph] = state.uniqueEmbeds[state.graph] || {};
     }
 
+    if(flags.embed === '@link' && id in link) {
+      // TODO: may want to also match an existing linked subject against
+      // the current frame ... so different frames could produce different
+      // subjects that are only shared in-memory when the frames are the same
+
+      // add existing linked subject
+      _addFrameOutput(parent, property, link[id]);
+      continue;
+    }
+
     // start output for subject
-    const output = {};
-    output['@id'] = id;
+    const output = {'@id': id};
     if(id.indexOf('_:') === 0) {
       util.addValue(state.bnodeMap, id, output, {propertyIsArray: true});
     }
-    state.link[id] = output;
+    link[id] = output;
+
+    // validate @embed
+    if((flags.embed === '@first' || flags.embed === '@last') && state.is11) {
+      throw new JsonLdError(
+        'Invalid JSON-LD syntax; invalid value of @embed.',
+        'jsonld.SyntaxError', {code: 'invalid @embed value', frame});
+    }
+
+    if(!state.embedded && state.uniqueEmbeds[state.graph].hasOwnProperty(id)) {
+      // skip adding this node object to the top level, as it was
+      // already included in another node object
+      continue;
+    }
 
     // if embed is @never or if a circular reference would be created by an
     // embed, the subject cannot be embedded, just add the reference;
     // note that a circular reference won't occur when the embed flag is
     // `@link` as the above check will short-circuit before reaching this point
-    if(flags.embed === '@never' ||
-      _createsCircularReference(subject, state.graph, state.subjectStack)) {
+    if(state.embedded &&
+      (flags.embed === '@never' ||
+      _createsCircularReference(subject, state.graph, state.subjectStack))) {
+      _addFrameOutput(parent, property, output);
+      continue;
+    }
+
+    // if only the first (or once) should be embedded
+    if(state.embedded &&
+       (flags.embed == '@first' || flags.embed == '@once') &&
+       state.uniqueEmbeds[state.graph].hasOwnProperty(id)) {
       _addFrameOutput(parent, property, output);
       continue;
     }
@@ -134,10 +164,10 @@ api.frame = (state, subjects, frame, parent, property = null) => {
       if(id in state.uniqueEmbeds[state.graph]) {
         _removeEmbed(state, id);
       }
-      state.uniqueEmbeds[state.graph][id] =
-        {parent, property};
     }
 
+    state.uniqueEmbeds[state.graph][id] = {parent, property};
+
     // push matching subject onto stack to enable circular embed checks
     state.subjectStack.push({subject, graph: state.graph});
 
@@ -150,23 +180,27 @@ api.frame = (state, subjects, frame, parent, property = null) => {
         subframe = {};
       } else {
         subframe = frame['@graph'][0];
+        recurse = !(id === '@merged' || id === '@default');
         if(!types.isObject(subframe)) {
           subframe = {};
         }
-        recurse = !(id === '@merged' || id === '@default');
       }
 
       if(recurse) {
-        state.graphStack.push(state.graph);
-        state.graph = id;
         // recurse into graph
         api.frame(
-          state,
+          {...state, graph: id, embedded: false},
           Object.keys(state.graphMap[id]).sort(), [subframe], output, '@graph');
-        state.graph = state.graphStack.pop;
       }
     }
 
+    // if frame has @included, recurse over its sub-frame
+    if('@included' in frame) {
+      api.frame(
+        {...state, embedded: false},
+        subjects, frame['@included'], output, '@included');
+    }
+
     // iterate over subject properties
     for(const prop of Object.keys(subject).sort()) {
       // copy keywords to output
@@ -191,36 +225,39 @@ api.frame = (state, subjects, frame, parent, property = null) => {
       }
 
       // add objects
-      for(let o of subject[prop]) {
+      for(const o of subject[prop]) {
         const subframe = (prop in frame ?
           frame[prop] : _createImplicitFrame(flags));
 
         // recurse into list
         if(graphTypes.isList(o)) {
+          const subframe =
+            (frame[prop] && frame[prop][0] && frame[prop][0]['@list']) ?
+              frame[prop][0]['@list'] :
+              _createImplicitFrame(flags);
+
           // add empty list
           const list = {'@list': []};
           _addFrameOutput(output, prop, list);
 
           // add list objects
           const src = o['@list'];
-          for(const n in src) {
-            o = src[n];
-            if(graphTypes.isSubjectReference(o)) {
-              const subframe = (prop in frame ?
-                frame[prop][0]['@list'] : _createImplicitFrame(flags));
+          for(const oo of src) {
+            if(graphTypes.isSubjectReference(oo)) {
               // recurse into subject reference
-              api.frame(state, [o['@id']], subframe, list, '@list');
+              api.frame(
+                {...state, embedded: true},
+                [oo['@id']], subframe, list, '@list');
             } else {
               // include other values automatically
-              _addFrameOutput(list, '@list', util.clone(o));
+              _addFrameOutput(list, '@list', util.clone(oo));
             }
           }
-          continue;
-        }
-
-        if(graphTypes.isSubjectReference(o)) {
+        } else if(graphTypes.isSubjectReference(o)) {
           // recurse into subject reference
-          api.frame(state, [o['@id']], subframe, output, prop);
+          api.frame(
+            {...state, embedded: true},
+            [o['@id']], subframe, output, prop);
         } else if(_valueMatch(subframe[0], o)) {
           // include other values, if they match
           _addFrameOutput(output, prop, util.clone(o));
@@ -231,7 +268,13 @@ api.frame = (state, subjects, frame, parent, property = null) => {
     // handle defaults
     for(const prop of Object.keys(frame).sort()) {
       // skip keywords
-      if(isKeyword(prop)) {
+      if(prop === '@type') {
+        if(!types.isObject(frame[prop][0]) ||
+           !('@default' in frame[prop][0])) {
+          continue;
+        }
+        // allow through default types
+      } else if(isKeyword(prop)) {
         continue;
       }
 
@@ -253,21 +296,20 @@ api.frame = (state, subjects, frame, parent, property = null) => {
 
     // if embed reverse values by finding nodes having this subject as a value
     // of the associated property
-    if('@reverse' in frame) {
-      for(const reverseProp of Object.keys(frame['@reverse']).sort()) {
-        const subframe = frame['@reverse'][reverseProp];
-        for(const subject of Object.keys(state.subjects)) {
-          const nodeValues =
-            util.getValues(state.subjects[subject], reverseProp);
-          if(nodeValues.some(v => v['@id'] === id)) {
-            // node has property referencing this subject, recurse
-            output['@reverse'] = output['@reverse'] || {};
-            util.addValue(
-              output['@reverse'], reverseProp, [], {propertyIsArray: true});
-            api.frame(
-              state, [subject], subframe, output['@reverse'][reverseProp],
-              property);
-          }
+    for(const reverseProp of Object.keys(frame['@reverse'] || {}).sort()) {
+      const subframe = frame['@reverse'][reverseProp];
+      for(const subject of Object.keys(state.subjects)) {
+        const nodeValues =
+          util.getValues(state.subjects[subject], reverseProp);
+        if(nodeValues.some(v => v['@id'] === id)) {
+          // node has property referencing this subject, recurse
+          output['@reverse'] = output['@reverse'] || {};
+          util.addValue(
+            output['@reverse'], reverseProp, [], {propertyIsArray: true});
+          api.frame(
+            {...state, embedded: true},
+            [subject], subframe, output['@reverse'][reverseProp],
+            property);
         }
       }
     }
@@ -280,6 +322,50 @@ api.frame = (state, subjects, frame, parent, property = null) => {
   }
 };
 
+/**
+ * Replace `@null` with `null`, removing it from arrays.
+ *
+ * @param input the framed, compacted output.
+ * @param options the framing options used.
+ *
+ * @return the resulting output.
+ */
+api.cleanupNull = (input, options) => {
+  // recurse through arrays
+  if(types.isArray(input)) {
+    const noNulls = input.map(v => api.cleanupNull(v, options));
+    return noNulls.filter(v => v); // removes nulls from array
+  }
+
+  if(input === '@null') {
+    return null;
+  }
+
+  if(types.isObject(input)) {
+    // handle in-memory linked nodes
+    if('@id' in input) {
+      const id = input['@id'];
+      if(options.link.hasOwnProperty(id)) {
+        const idx = options.link[id].indexOf(input);
+        if(idx !== -1) {
+          // already visited
+          return options.link[id][idx];
+        }
+        // prevent circular visitation
+        options.link[id].push(input);
+      } else {
+        // prevent circular visitation
+        options.link[id] = [input];
+      }
+    }
+
+    for(const key in input) {
+      input[key] = api.cleanupNull(input[key], options);
+    }
+  }
+  return input;
+};
+
 /**
  * Creates an implicit frame when recursing through subject matches. If
  * a frame doesn't have an explicit frame for a particular property, then
@@ -339,11 +425,14 @@ function _getFrameFlag(frame, options, name) {
     // true => "@last"
     // false => "@never"
     if(rval === true) {
-      rval = '@last';
+      rval = '@once';
     } else if(rval === false) {
       rval = '@never';
-    } else if(rval !== '@always' && rval !== '@never' && rval !== '@link') {
-      rval = '@last';
+    } else if(rval !== '@always' && rval !== '@never' && rval !== '@link' &&
+      rval !== '@first' && rval !== '@last' && rval !== '@once') {
+      throw new JsonLdError(
+        'Invalid JSON-LD syntax; invalid value of @embed.',
+        'jsonld.SyntaxError', {code: 'invalid @embed value', frame});
     }
   }
   return rval;
@@ -360,6 +449,30 @@ function _validateFrame(frame) {
       'Invalid JSON-LD syntax; a JSON-LD frame must be a single object.',
       'jsonld.SyntaxError', {frame});
   }
+
+  if('@id' in frame[0]) {
+    for(const id of util.asArray(frame[0]['@id'])) {
+      // @id must be wildcard or an IRI
+      if(!(types.isObject(id) || url.isAbsolute(id)) ||
+        (types.isString(id) && id.indexOf('_:') === 0)) {
+        throw new JsonLdError(
+          'Invalid JSON-LD syntax; invalid @id in frame.',
+          'jsonld.SyntaxError', {code: 'invalid frame', frame});
+      }
+    }
+  }
+
+  if('@type' in frame[0]) {
+    for(const type of util.asArray(frame[0]['@type'])) {
+      // @id must be wildcard or an IRI
+      if(!(types.isObject(type) || url.isAbsolute(type)) ||
+        (types.isString(type) && type.indexOf('_:') === 0)) {
+        throw new JsonLdError(
+          'Invalid JSON-LD syntax; invalid @type in frame.',
+          'jsonld.SyntaxError', {code: 'invalid frame', frame});
+      }
+    }
+  }
 }
 
 /**
@@ -412,104 +525,99 @@ function _filterSubject(state, subject, frame, flags) {
     const nodeValues = util.getValues(subject, key);
     const isEmpty = util.getValues(frame, key).length === 0;
 
-    if(isKeyword(key)) {
-      // skip non-@id and non-@type
-      if(key !== '@id' && key !== '@type') {
-        continue;
-      }
-      wildcard = false;
-
-      // check @id for a specific @id value
-      if(key === '@id') {
-        // if @id is not a wildcard and is not empty, then match or not on
-        // specific value
-        if(frame['@id'].length >= 0 && !types.isEmptyObject(frame['@id'][0])) {
-          return frame['@id'].includes(nodeValues[0]);
-        }
+    if(key === '@id') {
+      // match on no @id or any matching @id, including wildcard
+      if(types.isEmptyObject(frame['@id'][0] || {})) {
         matchThis = true;
-        continue;
+      } else if(frame['@id'].length >= 0) {
+        matchThis = frame['@id'].includes(nodeValues[0]);
       }
-
-      // check @type (object value means 'any' type, fall through to ducktyping)
-      if('@type' in frame) {
-        if(isEmpty) {
-          if(nodeValues.length > 0) {
-            // don't match on no @type
-            return false;
-          }
-          matchThis = true;
-        } else if(frame['@type'].length === 1 &&
-          types.isEmptyObject(frame['@type'][0])) {
-          // match on wildcard @type
-          matchThis = nodeValues.length > 0;
-        } else {
-          // match on a specific @type
-          for(const type of frame['@type']) {
-            if(nodeValues.some(tt => tt === type)) {
-              return true;
-            }
-          }
+      if(!flags.requireAll) {
+        return matchThis;
+      }
+    } else if(key === '@type') {
+      // check @type (object value means 'any' type,
+      // fall through to ducktyping)
+      wildcard = false;
+      if(isEmpty) {
+        if(nodeValues.length > 0) {
+          // don't match on no @type
           return false;
         }
+        matchThis = true;
+      } else if(frame['@type'].length === 1 &&
+        types.isEmptyObject(frame['@type'][0])) {
+        // match on wildcard @type if there is a type
+        matchThis = nodeValues.length > 0;
+      } else {
+        // match on a specific @type
+        for(const type of frame['@type']) {
+          if(types.isObject(type) && '@default' in type) {
+            // match on default object
+            matchThis = true;
+          } else {
+            matchThis = matchThis || nodeValues.some(tt => tt === type);
+          }
+        }
       }
-    }
-
-    // Forc a copy of this frame entry so it can be manipulated
-    const thisFrame = util.getValues(frame, key)[0];
-    let hasDefault = false;
-    if(thisFrame) {
-      _validateFrame([thisFrame]);
-      hasDefault = '@default' in thisFrame;
-    }
-
-    // no longer a wildcard pattern if frame has any non-keyword properties
-    wildcard = false;
-
-    // skip, but allow match if node has no value for property, and frame has a
-    // default value
-    if(nodeValues.length === 0 && hasDefault) {
+      if(!flags.requireAll) {
+        return matchThis;
+      }
+    } else if(isKeyword(key)) {
       continue;
-    }
+    } else {
+      // Force a copy of this frame entry so it can be manipulated
+      const thisFrame = util.getValues(frame, key)[0];
+      let hasDefault = false;
+      if(thisFrame) {
+        _validateFrame([thisFrame]);
+        hasDefault = '@default' in thisFrame;
+      }
 
-    // if frame value is empty, don't match if subject has any value
-    if(nodeValues.length > 0 && isEmpty) {
-      return false;
-    }
+      // no longer a wildcard pattern if frame has any non-keyword properties
+      wildcard = false;
 
-    if(thisFrame === undefined) {
-      // node does not match if values is not empty and the value of property
-      // in frame is match none.
-      if(nodeValues.length > 0) {
+      // skip, but allow match if node has no value for property, and frame has
+      // a default value
+      if(nodeValues.length === 0 && hasDefault) {
+        continue;
+      }
+
+      // if frame value is empty, don't match if subject has any value
+      if(nodeValues.length > 0 && isEmpty) {
         return false;
       }
-      matchThis = true;
-    } else if(types.isObject(thisFrame)) {
-      // node matches if values is not empty and the value of property in frame
-      // is wildcard
-      matchThis = nodeValues.length > 0;
-    } else {
-      if(graphTypes.isValue(thisFrame)) {
-        // match on any matching value
-        matchThis = nodeValues.some(nv => _valueMatch(thisFrame, nv));
-      } else if(graphTypes.isSubject(thisFrame) ||
-        graphTypes.isSubjectReference(thisFrame)) {
-        matchThis =
-          nodeValues.some(nv => _nodeMatch(state, thisFrame, nv, flags));
-      } else if(graphTypes.isList(thisFrame)) {
-        const listValue = thisFrame['@list'][0];
-        if(graphTypes.isList(nodeValues[0])) {
-          const nodeListValues = nodeValues[0]['@list'];
 
-          if(graphTypes.isValue(listValue)) {
-            // match on any matching value
-            matchThis = nodeListValues.some(lv => _valueMatch(listValue, lv));
-          } else if(graphTypes.isSubject(listValue) ||
-            graphTypes.isSubjectReference(listValue)) {
-            matchThis = nodeListValues.some(lv => _nodeMatch(
-              state, listValue, lv, flags));
+      if(thisFrame === undefined) {
+        // node does not match if values is not empty and the value of property
+        // in frame is match none.
+        if(nodeValues.length > 0) {
+          return false;
+        }
+        matchThis = true;
+      } else {
+        if(graphTypes.isList(thisFrame)) {
+          const listValue = thisFrame['@list'][0];
+          if(graphTypes.isList(nodeValues[0])) {
+            const nodeListValues = nodeValues[0]['@list'];
+
+            if(graphTypes.isValue(listValue)) {
+              // match on any matching value
+              matchThis = nodeListValues.some(lv => _valueMatch(listValue, lv));
+            } else if(graphTypes.isSubject(listValue) ||
+              graphTypes.isSubjectReference(listValue)) {
+              matchThis = nodeListValues.some(lv => _nodeMatch(
+                state, listValue, lv, flags));
+            }
           }
+        } else if(graphTypes.isValue(thisFrame)) {
+          matchThis = nodeValues.some(nv => _valueMatch(thisFrame, nv));
+        } else if(graphTypes.isSubjectReference(thisFrame)) {
+          matchThis =
+            nodeValues.some(nv => _nodeMatch(state, thisFrame, nv, flags));
+        } else if(types.isObject(thisFrame)) {
+          matchThis = nodeValues.length > 0;
         } else {
-          // value must be a list to match
           matchThis = false;
         }
       }
@@ -574,6 +682,68 @@ function _removeEmbed(state, id) {
   removeDependents(id);
 }
 
+/**
+ * Removes the @preserve keywords from expanded result of framing.
+ *
+ * @param input the framed, framed output.
+ * @param options the framing options used.
+ *
+ * @return the resulting output.
+ */
+function _cleanupPreserve(input, options) {
+  // recurse through arrays
+  if(types.isArray(input)) {
+    return input.map(value => _cleanupPreserve(value, options));
+  }
+
+  if(types.isObject(input)) {
+    // remove @preserve
+    if('@preserve' in input) {
+      return input['@preserve'][0];
+    }
+
+    // skip @values
+    if(graphTypes.isValue(input)) {
+      return input;
+    }
+
+    // recurse through @lists
+    if(graphTypes.isList(input)) {
+      input['@list'] = _cleanupPreserve(input['@list'], options);
+      return input;
+    }
+
+    // handle in-memory linked nodes
+    if('@id' in input) {
+      const id = input['@id'];
+      if(options.link.hasOwnProperty(id)) {
+        const idx = options.link[id].indexOf(input);
+        if(idx !== -1) {
+          // already visited
+          return options.link[id][idx];
+        }
+        // prevent circular visitation
+        options.link[id].push(input);
+      } else {
+        // prevent circular visitation
+        options.link[id] = [input];
+      }
+    }
+
+    // recurse through properties
+    for(const prop in input) {
+      // potentially remove the id, if it is an unreference bnode
+      if(prop === '@id' && options.bnodesToClear.includes(input[prop])) {
+        delete input['@id'];
+        continue;
+      }
+
+      input[prop] = _cleanupPreserve(input[prop], options);
+    }
+  }
+  return input;
+}
+
 /**
  * Adds framing output to the given parent.
  *
