@@ -40,7 +40,6 @@ const IdentifierIssuer = util.IdentifierIssuer;
 const JsonLdError = require('./JsonLdError');
 const LRU = require('lru-cache');
 const NQuads = require('./NQuads');
-const Rdfa = require('./Rdfa');
 
 const {expand: _expand} = require('./expand');
 const {flatten: _flatten} = require('./flatten');
@@ -989,9 +988,6 @@ jsonld.unregisterRDFParser = function(contentType) {
 jsonld.registerRDFParser('application/n-quads', NQuads.parse);
 jsonld.registerRDFParser('application/nquads', NQuads.parse);
 
-// register the RDFa API RDF parser
-jsonld.registerRDFParser('rdfa-api', Rdfa.parse);
-
 /* URL API */
 jsonld.url = require('./url');
 
