@@ -6,6 +6,7 @@
 const {createNodeMap} = require('./nodeMap');
 const {isKeyword} = require('./context');
 const graphTypes = require('./graphTypes');
+const jsonCanonicalize = require('canonicalize');
 const types = require('./types');
 const util = require('./util');
 
@@ -18,6 +19,7 @@ const {
   RDF_TYPE,
   // RDF_PLAIN_LITERAL,
   // RDF_XML_LITERAL,
+  RDF_JSON_LITERAL,
   // RDF_OBJECT,
   RDF_LANGSTRING,
 
@@ -85,8 +87,7 @@ api.toRDF = (input, options) => {
  */
 function _graphToRDF(dataset, graph, graphTerm, issuer, options) {
   const ids = Object.keys(graph).sort();
-  for(let i = 0; i < ids.length; ++i) {
-    const id = ids[i];
+  for(const id of ids) {
     const node = graph[id];
     const properties = Object.keys(node).sort();
     for(let property of properties) {
@@ -126,22 +127,17 @@ function _graphToRDF(dataset, graph, graphTerm, issuer, options) {
           continue;
         }
 
-        // convert @list to triples
-        if(graphTypes.isList(item)) {
-          _listToRDF(
-            item['@list'], issuer, subject, predicate, dataset, graphTerm);
-        } else {
-          // convert value or node object to triple
-          const object = _objectToRDF(item);
-          // skip null objects (they are relative IRIs)
-          if(object) {
-            dataset.push({
-              subject,
-              predicate,
-              object,
-              graph: graphTerm
-            });
-          }
+        // convert list, value or node object to triple
+        const object =
+          _objectToRDF(item, issuer, dataset, graphTerm, options.rdfDirection);
+        // skip null objects (they are relative IRIs)
+        if(object) {
+          dataset.push({
+            subject,
+            predicate,
+            object,
+            graph: graphTerm
+          });
         }
       }
     }
@@ -154,59 +150,71 @@ function _graphToRDF(dataset, graph, graphTerm, issuer, options) {
  *
  * @param list the @list value.
  * @param issuer a IdentifierIssuer for assigning blank node names.
- * @param subject the subject for the head of the list.
- * @param predicate the predicate for the head of the list.
  * @param dataset the array of quads to append to.
  * @param graphTerm the graph term for each quad.
+ *
+ * @return the head of the list.
  */
-function _listToRDF(list, issuer, subject, predicate, dataset, graphTerm) {
+function _listToRDF(list, issuer, dataset, graphTerm, rdfDirection) {
   const first = {termType: 'NamedNode', value: RDF_FIRST};
   const rest = {termType: 'NamedNode', value: RDF_REST};
   const nil = {termType: 'NamedNode', value: RDF_NIL};
 
+  const last = list.pop();
+  // Result is the head of the list
+  const result = last ? {termType: 'BlankNode', value: issuer.getId()} : nil;
+  let subject = result;
+
   for(const item of list) {
-    const blankNode = {termType: 'BlankNode', value: issuer.getId()};
+    const object = _objectToRDF(item, issuer, dataset, graphTerm, rdfDirection);
+    const next = {termType: 'BlankNode', value: issuer.getId()};
     dataset.push({
       subject,
-      predicate,
-      object: blankNode,
+      predicate: first,
+      object,
       graph: graphTerm
     });
-
-    subject = blankNode;
-    predicate = first;
-    const object = _objectToRDF(item);
-
-    // skip null objects (they are relative IRIs)
-    if(object) {
-      dataset.push({
-        subject,
-        predicate,
-        object,
-        graph: graphTerm
-      });
-    }
-
-    predicate = rest;
+    dataset.push({
+      subject,
+      predicate: rest,
+      object: next,
+      graph: graphTerm
+    });
+    subject = next;
   }
 
-  dataset.push({
-    subject,
-    predicate,
-    object: nil,
-    graph: graphTerm
-  });
+  // Tail of list
+  if(last) {
+    const object = _objectToRDF(last, issuer, dataset, graphTerm, rdfDirection);
+    dataset.push({
+      subject,
+      predicate: first,
+      object,
+      graph: graphTerm
+    });
+    dataset.push({
+      subject,
+      predicate: rest,
+      object: nil,
+      graph: graphTerm
+    });
+  }
+
+  return result;
 }
 
 /**
- * Converts a JSON-LD value object to an RDF literal or a JSON-LD string or
- * node object to an RDF resource.
+ * Converts a JSON-LD value object to an RDF literal or a JSON-LD string,
+ * node object to an RDF resource, or adds a list.
  *
  * @param item the JSON-LD value or node object.
+ * @param issuer a IdentifierIssuer for assigning blank node names.
+ * @param dataset the dataset to append RDF quads to.
+ * @param graphTerm the graph term for each quad.
  *
  * @return the RDF literal or RDF resource.
  */
-function _objectToRDF(item) {
+function _objectToRDF(item, issuer, dataset, graphTerm, rdfDirection) {
   const object = {};
 
   // convert value object to RDF
@@ -219,8 +227,11 @@ function _objectToRDF(item) {
     let value = item['@value'];
     const datatype = item['@type'] || null;
 
-    // convert to XSD datatypes as appropriate
-    if(types.isBoolean(value)) {
+    // convert to XSD/JSON datatypes as appropriate
+    if(datatype === '@json') {
+      object.value = jsonCanonicalize(value);
+      object.datatype.value = RDF_JSON_LITERAL;
+    } else if(types.isBoolean(value)) {
       object.value = value.toString();
       object.datatype.value = datatype || XSD_BOOLEAN;
     } else if(types.isDouble(value) || datatype === XSD_DOUBLE) {
@@ -233,6 +244,13 @@ function _objectToRDF(item) {
     } else if(types.isNumber(value)) {
       object.value = value.toFixed(0);
       object.datatype.value = datatype || XSD_INTEGER;
+    } else if(rdfDirection === 'i18n-datatype' &&
+      '@direction' in item) {
+      const datatype = 'https://www.w3.org/ns/i18n#' +
+        (item['@language'] || '') +
+        `_${item['@direction']}`;
+      object.datatype.value = datatype;
+      object.value = value;
     } else if('@language' in item) {
       object.value = value;
       object.datatype.value = datatype || RDF_LANGSTRING;
@@ -241,6 +259,11 @@ function _objectToRDF(item) {
       object.value = value;
       object.datatype.value = datatype || XSD_STRING;
     }
+  } else if(graphTypes.isList(item)) {
+    const _list =
+      _listToRDF(item['@list'], issuer, dataset, graphTerm, rdfDirection);
+    object.termType = _list.termType;
+    object.value = _list.value;
   } else {
     // convert string/node object to RDF
     const id = types.isObject(item) ? item['@id'] : item;
