@@ -5,8 +5,12 @@
  *
  * Copyright (c) 2011-2017 Digital Bazaar, Inc. All rights reserved.
  */
+/* global IdlArray */
+/* global add_result_callback */
+/* global add_completion_callback */
+/* global done */
+/* eslint-disable no-var */
 const assert = require('chai').assert;
-const jsonld = require('..');
 
 require('./webidl/testharness.js');
 require('./webidl/WebIDLParser.js');
@@ -66,11 +70,17 @@ return new Promise((resolve, reject) => {
         'expected "JsonLdProcessor" but got "t"') !== -1) {
         this.skip();
       }
+      if(msg.indexOf(
+        'wrong value for JsonLdProcessor.name ' +
+        'expected "JsonLdProcessor" but got "e"') !== -1) {
+        this.skip();
+      }
       //earl.addAssertion({'@id': ?}, test.status === 0);
       assert.equal(test.status, 0, test.message);
       done();
     };
   });
+  /* eslint-disable-next-line no-unused-vars */
   add_completion_callback(function(tests, status) {
     resolve(suite);
   });
